/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


////////////////////////////////////////////////////////////////////////////
//
// CompositePart Methods
//
////////////////////////////////////////////////////////////////////////////

#include <libc.h>
#include <stdio.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "VarParams.h"
#include "IntBtree.h"
#include "StringBtree.h"
#include "baidlist.h"
#include <E/sm_btree.h>

extern int	nextAtomicId;
extern int	debugMode;
extern char*	types[NumTypes];
extern persistent btree<int> BuildDateIdx;

extern BAIdList* private_cp;
extern BAIdList* shared_cp;


////////////////////////////////////////////////////////////////////////////
//
// CompositePart Constructor
//
////////////////////////////////////////////////////////////////////////////

CompositePart::CompositePart(dbint cpId, dbint bulkLoadFlag)
{
    int 	partDate;
    int 	i, from, to;
    int 	atomicid, length;
    AtomicPart*	atomicParts[NumAtomicPerComp];
    Connection*	cn;
    BaseAssembly* ba;
    char	title[TitleSize];

    if (debugMode) {
        printf("CompositePart::CompositePart(cpId = %d)\n", cpId);
    }

    // initialize the simple stuff
    id = cpId;
    int typeNo = (int) (random() % NumTypes);
    strncpy(type, types[typeNo], TypeSize);

    // for the build date, decide if this part is young or old, and then
    // randomly choose a date in the required range

    if (cpId % YoungCompFrac == 0) {
        // young one
        if (debugMode) {
            fprintf(stderr, "(young composite part, id = %d.)\n", id);
        }
        buildDate = MinYoungCompDate +
                    (int)(random() % (MaxYoungCompDate - MinYoungCompDate + 1));
    } else {
        // old one
        if (debugMode) {
            fprintf(stderr, "(old composite part, id = %d.)\n", id);
        }
        buildDate = MinOldCompDate +
                    (int)(random() % (MaxOldCompDate - MinOldCompDate + 1));
    }

    parts = new (CompPartSet) Assoc(&CompPartSet);
    usedInPriv = new (CompPartSet) Assoc(&CompPartSet);
    usedInShar = new (CompPartSet) Assoc(&CompPartSet);

    // initialize the documentation (indexed by its title and id) ...
    sprintf(title, "Composite Part %08d", cpId);
    documentation = new (DocumentSet) Document(cpId, this);

    // insert title into document index
    DocumentIdx->insert(title, documentation);

    // insert id into document id index
    DocumentIdIdx->insert(cpId, documentation);

    // now create the atomic parts (indexed by their ids) ...
    for (i = 0; i < NumAtomicPerComp; i++) 
    {
	atomicid = nextAtomicId + i;

	// create a new atomic part
	// the AtomicPart constructor takes care of setting up
	// the back pointer to the containing CompositePart()
	// atomicParts[i] = new (CompPartSet) AtomicPart(atomicid,this);
	// now in its own collection
	atomicParts[i] = new (AtomicPartSet) AtomicPart(atomicid,this);

	// if not bulkLoading, update the two indices on atomic parts
	if (bulkLoadFlag == FALSE)
	{
	    // stick the id of the part into the Atomic Part index
	    AtomicPartIdx->insert(atomicid, atomicParts[i]);
	    // add the buildDate to the BuildDateIndex
	    BuildDateIdx.insert(atomicParts[i]->buildDate, atomicParts[i]);
	}

	// connect the atomic part to the composite part
	parts->add(atomicParts[i]);

	// first atomic part is the root part
	if (i == 0) rootPart = atomicParts[i];
    }

    // ... and then wire them semi-randomly together (as a ring plus random
    // additional connections to ensure full part reachability for traversals)

    for (from = 0; from < NumAtomicPerComp; from++) {
	for (i = 0; i < NumConnPerAtomic; i++) {
	    if (i == 0) {
	        to = (from + 1) % NumAtomicPerComp;
	    } else {
	        to = (int) (random() % NumAtomicPerComp);
	    }
	    cn = new (CompPartSet) Connection(atomicParts[from], 
			atomicParts[to]);
	}
    }
    nextAtomicId += NumAtomicPerComp;

    // finally insert this composite part as a child of the base
    // assemblies that use it

    // first the assemblies using the comp part as a shared component
    shared_cp[cpId].initScan();

    // get the first base assembly
    ba = shared_cp[cpId].next();
    while (ba != NULL)
    {
	// add this assembly to the list of assemblies in which
	// this composite part is used as a shared member
	(this->usedInShar)->add(ba);

	// then add the composite part cp to the list of shared parts used
	// in this assembly
	(ba->componentsShar)->add(this);

	// continue with next base assembly
        ba = shared_cp[cpId].next();
    }

    // next the assemblies using the comp part as a private component
    private_cp[cpId].initScan();

    // get the first base assembly
    ba = private_cp[cpId].next();
    while (ba != NULL)
    {
	// first add this assembly to the list of assemblies in which
	// this composite part is used as a private member
	(this->usedInPriv)->add(ba);

	// then add the composite part cp to the list of private parts used
	// in the assembly
	(ba->componentsPriv)->add(this);

	// continue with next base assembly
        ba = private_cp[cpId].next();
    }
}

// destructor
CompositePart::~CompositePart()
{
    AssocItr*     baseI;  // iterator for traversing subassemblies
    BaseAssembly* baH;
    AssocItr*     atomI;  // iterator for traversing atomic parts
    AtomicPart*	  apH;
    AtomicPart*	  apH2;
    Document* 	docH;
    char	docTitle[40];

    strcpy(docTitle,documentation->title);
    docH = (Document*) DocumentIdx->remove(docTitle);
    if (docH != documentation)
      printf("error deleting document from title index in CompPart destructor\n");
    docH = (Document*) DocumentIdIdx->remove(documentation->id);
    if (docH != documentation)
      printf("error deleting document from id index in CompPart destructor\n");
    
    delete documentation; 

    // walk through usedInPriv association deleting all backward references
    // to this composite part from base assemblies of which it is a part
    baseI = new AssocItr(usedInPriv);  // establish baseassembly iterator
    baH = (BaseAssembly*) baseI->next();
    while (baH != NULL)
    {
	baH->componentsPriv->remove(this);
    	baH = (BaseAssembly*) baseI->next();
    }
    delete usedInPriv; // delete the association object
    delete baseI;

    // walk through usedInShar association deleting all backward references
    // to this composite part from base assemblies of which it is a part
    baseI = new AssocItr(usedInShar);  // establish baseassembly iterator
    baH = (BaseAssembly*) baseI->next();
    while (baH != NULL)
    {
	baH->componentsShar->remove(this);
    	baH = (BaseAssembly*) baseI->next();
    }
    delete usedInShar; // delete the association object
    delete baseI;

    // delete all atomic parts that compose this composite part
    atomI = new AssocItr(parts);  // establish atomic part iterator
    apH = (AtomicPart*) atomI->next();
    while (apH != NULL)
    {
	apH2 = (AtomicPart*) AtomicPartIdx->remove(apH->id);
	if (apH != apH2) printf("index error when deleting atomic parts \n");
	delete apH; // the atomic part destructor cleans up
		    // connection objects
    	apH = (AtomicPart*) atomI->next();
    }
    delete parts;  // delete parts association
    delete atomI;

    // dts 3-11-93 AllCompParts is not used, so don't maintain it.
    // (used in Reorg code though, but that's not called currently)
    // AllCompParts.remove(this);  // finally remove from the extent
}
