/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <E/trans.h>
#include <E/collection.h>
#include <E/dbStrings.h>
#include "PartIdSet.h"

//--------------------------------------------------------------------------
// Start with some necessary preliminaries
//--------------------------------------------------------------------------

#define TypeSize	10
#define TitleSize	40

// base size for associations
#define BaseSize	10 // tried 4, but database blew up.
// #define OverflowSize	450
#define OverflowSize	BaseSize

typedef enum {Complex, Base} AssemblyType;

typedef enum { Trav1, Trav1WW, Trav2a, Trav2b, Trav2c, Trav3a, Trav3b, 
	Trav3c, Trav4, Trav5do, Trav5undo, Trav6, Trav7, Trav8, Trav9,
	Trav10,
       Query1, Query2, Query3, Query4, Query5, Query6, Query7, Query8, 
       Insert, Delete, Reorg1, Reorg2, WarmUpdate, MultiTrav1, MultiTrav2,
       MultiTrav3, MultiTrav4, MultiTrav5, MultiTrav6 
      } BenchmarkOp;

typedef enum {UpdateOne, UpdateAll, UpdateRepeat} UpdateType;
typedef enum {UpdateDirectionDo, UpdateDirectionUndo} UpdateDirectionType;

// association class. should be made into a generic
dbclass AssocBase {
friend class AssocItr;
public:
    collection <dbvoid>*   assocFile;	// place to stick the association
    dbint		   allocated;	// actual size
    dbint		   curSize;	// number of elements used 
    dbint		   freePtr;	// index into members array
    dbclass AssocOverflow* overflow;	// beginning of the overflow chain
    dbvoid*	members[1]; // connection pointers
    // there will actually be more space in Assoc & AssocOveflow.
    // rewrite as non-virual variable size.  The only
    // differenct between Assoc and AssocOverflow is size of Members.
    dbvoid add(dbvoid* member);
    dbvoid remove(dbvoid* member);
    inline dbvoid* member(int i) { return members[i]; }
    ~AssocBase(); // don't seem to work =0;
	
};

// association class - for small sets
dbclass Assoc: public AssocBase {
private:
    dbvoid*	members_sp[BaseSize-1]; // connection pointers
public:
    Assoc(collection<dbvoid>* cpt);
};

// overflow block - used for large sets
dbclass AssocOverflow: public AssocBase {
private:
    dbvoid*	members_sp[OverflowSize-1]; // connection pointers
public:
    AssocOverflow(collection<dbvoid>* cpt);
};


// association iteration class. should be made into a generic
class AssocItr {
private:
    AssocBase*	current;	// pointer to current association object 
    dbint	scanPtr;	// index into members array of current object
public:
    AssocItr(Assoc* assoc);
    ~AssocItr() { };
    dbvoid* next();	// returns next object in scan
};
	


//--------------------------------------------------------------------------
// AtomicPart objects are the primitives for building up designs
//	- modeled after the Sun/OO1 benchmark's parts
//--------------------------------------------------------------------------

dbclass AtomicPart {
public:
    dbint	    id;
    dbchar	    type[TypeSize];
    dbint	    buildDate;
    dbint	    x, y;
    dbint             docId;
    dbclass Assoc*    to;    // to connection objects
    dbclass Assoc*    from;  // back pointers
    dbclass CompositePart*  partOf;  // up pointer

    AtomicPart(dbint ptId, CompositePart*  cp);
    ~AtomicPart();
    dbvoid swapXY();
    dbvoid toggleDate();
    dbvoid DoNothing();

    dbint traverse(BenchmarkOp op, PartIdSet& visitedIds);
};

//--------------------------------------------------------------------------
// Connection objects are used to wire AtomicParts together
//	- similarly, modeled after Sun/OO1 connections
//--------------------------------------------------------------------------

dbclass Connection {
public:
    dbchar		type[TypeSize];
    dbint		length;
    dbclass AtomicPart*	from;
    dbclass AtomicPart* to;

    Connection(AtomicPart* fromPart, AtomicPart* toPart);
    ~Connection(){};
};


//--------------------------------------------------------------------------
// CompositeParts are parts constructed from AtomicParts
//	- entry in a library of reusesable components
//	- implementation is a graph of atomic parts
//	- provides unit of significant access locality
//	- each has an associated (unique) document object
//--------------------------------------------------------------------------

dbclass CompositePart {
public:
    dbint		id;
    dbchar		type[TypeSize];
    dbint		buildDate;
    dbclass Document*	documentation;
    dbclass Assoc*	parts;
    dbclass AtomicPart*	rootPart;
    // list of assemblies in which part is used as a private component
    Assoc*		usedInPriv;
    // list of assemblies in which part is used as a shared component
    Assoc*		usedInShar;

    CompositePart(dbint cpId, dbint bulkLoadFlag);
    ~CompositePart();
    dbint traverse(BenchmarkOp op);
    dbint traverse7();
    dbint reorg1();
    dbint reorg2();
};


//--------------------------------------------------------------------------
// Document objects are used to hold a description of some particular
// CompositePart object
//--------------------------------------------------------------------------

dbclass Document {
public:
    dbchar			title[TitleSize];
    dbint                       id;
    dbchar*			text;
    dbint			textSize; // actual size of text area
    dbint			textLen;  // amount actually being used
    dbclass CompositePart*	part;

    Document(dbint cpId, CompositePart* part);
    ~Document();
    dbint searchText(char c);
    dbint replaceText(char *oldString, char* newString);
};

//--------------------------------------------------------------------------
// Manual objects are used to hold a description of some particular
// module --- essentially just large documents.
//--------------------------------------------------------------------------

dbclass Manual {
public:
    dbchar			title[TitleSize];
    dbint                       id;
    dbchar*			text;
    dbint			textSize; // actual size of text area
    dbint			textLen;  // amount actually being used
    dbclass Module*	        mod;

    Manual(dbint modId, Module* myModule);
    ~Manual();
    dbint searchText(char c);
    dbint replaceText(char *oldString, char* newString);
    dbint firstLast();
};

//--------------------------------------------------------------------------
// Assembly objects are design instances built up recursively from
// from other Assembly objects and (at the leaves only) CompositeParts
//	- hierarchical (tree) structure for designs
//	- may share composite parts with other assemblies
//	- nonleaf and leaf assembly subtypes
//--------------------------------------------------------------------------

dbclass Assembly {
public:
    dbint			id;
    dbchar			type[TypeSize];
    dbint			buildDate;
    AssemblyType		assmType;
    dbclass ComplexAssembly*	superAssembly;

    virtual dbint traverse(BenchmarkOp op) = 0;
    virtual dbint traverse7(PartIdSet& visitedComplexIds) = 0;
    virtual AssemblyType mytype() = 0;
    dbvoid DoNothing();
};


dbclass ComplexAssembly: public Assembly {
public:
    dbclass Assoc*	subAssemblies;  // list of composite subassemblies

    ComplexAssembly(dbint asId, dbclass Module* module, 
	ComplexAssembly* parentAssembly, dbint levelNo);

    dbint traverse(BenchmarkOp op);
    dbint traverse7(PartIdSet& visitedComplexIds);
    virtual AssemblyType mytype() { return Complex;}
};


dbclass BaseAssembly: public Assembly {
public:
    dbclass Assoc*	componentsPriv; // list of private composite parts
    dbclass Assoc*	componentsShar; // list of shared composite parts

    BaseAssembly(dbint asId, dbclass Module* module,
    	             ComplexAssembly* parentAssembly);
    dbint traverse(BenchmarkOp op);
    dbint traverse7(PartIdSet& visitedComplexIds);
    virtual AssemblyType mytype() { return Base;}
};


//--------------------------------------------------------------------------
// Modules are the designs resulting from Assembly composition
//	- unit of scaleup for the benchmark database
//	- may share composite parts with other modules
//--------------------------------------------------------------------------

dbclass Module {
public:
    dbint			id;
    dbchar			type[TypeSize];
    dbint			buildDate;
    dbclass ComplexAssembly*	designRoot;
    dbclass Manual*	        man;  // manual object
    Assoc*			allBases;  // all base assemblies in a module

    Module(dbint modId);
    ~Module();
    dbint traverse(BenchmarkOp op);
    dbint scanManual();
    dbint firstLast();
};

